// (c) Copyright 2001-2002 iwombat.com all rights reserved
package com.iwombat.foundation.persist;

import java.util.HashMap;
import java.util.Iterator;
import java.util.ArrayList;

import com.iwombat.foundation.Identifier;
import com.iwombat.foundation.IdentifierFactory;

/**
 * Coordinates and manages transaction instances
 *
 * @author  $Author: brentp $
 * @version $Revision: 1.2 $
 * 
 * <br>
 * <dl><dt><b>Date:</b><dd>$Date: 2002/11/18 16:04:40 $</dd></dl>
 */

public class TransactionCoordinator {

    private static long DEFAULT_TTL_MILLIS=30000;
    private static HashMap transactionInstances= new HashMap();
    
    private static HashMap getTransactionInstances(){
        return transactionInstances;
    }
    
    /**
     * aborts the specified transaction
     *
     * @param the transaction token representing the transaction to abort
     */
    public static void abortTransaction(TransactionToken transactionToken){

        HashMap activeTransactionInstances = getTransactionInstances();
        try{
            TransactionInstance instance=(TransactionInstance)activeTransactionInstances.get(transactionToken);
            instance.abort();
        }
        catch(Throwable t){
            throw new PersistenceException("Exception caught while aborting transaction",t);
        }
        finally{
            activeTransactionInstances.remove(transactionToken.getTokenId());
        }
    }
    
    /**
     * commits the specified transaction
     *
     * @param the transaction token representing the transaction to commit
     */
    public static void commitTransaction(TransactionToken transactionToken){
        
        HashMap activeTransactionInstances = getTransactionInstances();
        try{
            TransactionInstance instance=(TransactionInstance)activeTransactionInstances.get(transactionToken);
            instance.commit();
        }
        catch(Throwable t){
            throw new PersistenceException("Exception caught while commiting transaction",t);
        }
        finally{
            activeTransactionInstances.remove(transactionToken.getTokenId());
        }
        
    }
    
    /**
     * starts a new transaction
     *
     * @param the transaction token representing the transaction to start
     * @param the transaction instance representing the new transaction
     */
    protected static void startTransaction(TransactionToken transactionToken, TransactionInstance transactionInstance){
        
        HashMap activeTransactionInstances = getTransactionInstances();
        
        if (transactionToken==null) throw new IllegalArgumentException("transaction token specified was null!");
        if (transactionInstance==null) throw new IllegalArgumentException("transaction instance specified was null!");
        
        activeTransactionInstances.put(transactionToken,transactionInstance);
    }
    
    /**
     * creates a new transaction token with a specified minimum time to live
     *
     * @param the minimum time this transaction will live, before it times out, in milliseconds
     */
    public static TransactionToken createTransactionToken(long timeToLiveMillis){
    
        //clean up any timed-out transactions that have been orphaned by poorly written client code
        cleanUpOrphanedTransactions();
        
        TransactionToken token=new TransactionToken(IdentifierFactory.createUniqueIdentifier(),timeToLiveMillis);

        return token;
    }
    
    /**
     * creates a new transaction token with the default minimum time to live
     *
     */
    public static TransactionToken createTransactionToken(){
        
        return createTransactionToken(TransactionCoordinator.DEFAULT_TTL_MILLIS);
    }
    
    /**
     * retrieves an active transaction instance
     *
     * @param the transaction token representing the transaction instance to retrieve
     */
    protected static TransactionInstance getTransactionInstance(TransactionToken transactionToken){
        
        return (TransactionInstance)getTransactionInstances().get(transactionToken);
    }
 

    /**
     * cleans up orphaned transactions
     *
     */
    private synchronized static void cleanUpOrphanedTransactions(){
        

        HashMap activeTransactionInstances = getTransactionInstances();
        TransactionToken transactionToken;
        TransactionInstance transactionInstance;
        
        ArrayList deadTransactionTokens=null;

        for (Iterator i=activeTransactionInstances.keySet().iterator();i.hasNext();){
         
            transactionToken=(TransactionToken)i.next();
            
            if (transactionToken.isExpired()){
                try{
                    transactionInstance=(TransactionInstance)activeTransactionInstances.get(transactionToken);
                    transactionInstance.abort();
                }
                catch(Throwable t){} //swallow any errors
                
                finally{
                    if (deadTransactionTokens==null) deadTransactionTokens=new ArrayList();
                    
                    //save them up for later removal
                    deadTransactionTokens.add(transactionToken);
                }
            }
            
        }
        
        //remove any dead transaction tokens
        if (deadTransactionTokens!=null){
            
            for (Iterator i=deadTransactionTokens.iterator();i.hasNext();){
                activeTransactionInstances.remove(i.next());
            }
            
        }
    }
    
}
